package Anarres::Mud::Driver::Program::Type;

use strict;
use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS %PROMOTIONS);
use Exporter;
use Carp qw(:DEFAULT cluck);
use Data::Dumper;

BEGIN {
	no strict qw(refs);

	my %types = (
		T_VOID		=> "V",
		T_NIL		=> "N",
		T_MIXED		=> "_",
		T_BOOL		=> "b",
		T_CLOSURE	=> "f",
		T_INTEGER	=> "i",
		T_MAPPING	=> "m",
		T_OBJECT	=> "o",
		T_STRING	=> "s",
		T_ARRAY		=> "P_",
			);
	my $package = __PACKAGE__;

	foreach (keys %types) {
		# Why doesn't this work to let me reference the variable
		# as \$$_ in the eval below?
		# *{"${package}::$_"} = \${"${package}::$_"};		# "use vars"


		my $tpack = lc $_;
		$tpack =~ s/^T_//i;
		$tpack = ucfirst $tpack;

		# XXX I don't really need the $tpack stuff. I'm not going
		# to recurse over the type tree, it's too slow.

		eval qq{
			 *${package}::$_ = sub () { \$${package}::$_; };
			\@${package}::${tpack}::ISA = qw($package);
			\$${package}::$_ =
				bless \\\${types{$_}}, "${package}::${tpack}";
		}; die $@ if $@;

			# *${package}::$_ = \\\$${package}::$_;
			# my \$self = "$types{$_}";
			# return bless \\\$self, "$package";

		# print "$_: ${$_} = $${$_}\n";
	}

	# print Dumper(\%{"${package}::"});

	@ISA = qw(Exporter);
	@EXPORT_OK = keys %types;
	%EXPORT_TAGS = ( types => \@EXPORT_OK );

	# XXX Of course, the identity isn't _usually_ the correct promotion.
	my $identity = sub { $_[0] };
	# XXX Insert an assert node...
	my $assert = $identity;

	# The < sign never appears in a type string.

	%PROMOTIONS = (
		$types{T_INTEGER}	. '<' . $types{T_BOOL}		=> $identity,
		$types{T_ARRAY}		. '<' . $types{T_BOOL}		=> $identity,
		$types{T_MAPPING}	. '<' . $types{T_BOOL}		=> $identity,
		$types{T_OBJECT}	. '<' . $types{T_BOOL}		=> $identity,
		$types{T_STRING}	. '<' . $types{T_BOOL}		=> $identity,
		$types{T_CLOSURE}	. '<' . $types{T_BOOL}		=> $identity,

		# Should this have special treatment?
		$types{T_NIL}		. '<' . $types{T_BOOL}		=> $identity,

		# This one should actually work...
		$types{T_BOOL}		. '<' . $types{T_INTEGER}	=> $identity,

		# Do I want this? (for addition, hash deref, assignment)
		$types{T_INTEGER}	. '<' . $types{T_STRING}	=> $identity,
		# Or this? (for hash deref)
		$types{T_OBJECT}	. '<' . $types{T_STRING}	=> $identity,
			);

	# In increasing priority order - we overwrite.
	foreach (values %types) {
		# Assert to turn a mixed into anything
		$PROMOTIONS{$types{T_MIXED} . '<' . $_} = $assert;		# OK

		# Anything can become a mixed
		$PROMOTIONS{$_ . '<' . $types{T_MIXED}} = $identity;	# OK

		# Nil can become anything
		$PROMOTIONS{$types{T_NIL} . '<' . $_} = $identity;	# ::False?

		# Anything is itself
		$PROMOTIONS{$_ . '<' . $_} = $identity;	# overwrites "_<_"
	}

#	print "Available types are\n" .
#						join("\n", sort keys %types) . "\n";;
#	print "Available promotions are\n" .
#						join("\n", sort keys %PROMOTIONS) . "\n";;
}

sub struct {
	my ($class, @fields) = @_;			# @fields are Types
	my $self = "S{" . join(",", map { $$_ } @fields) . "}";
	return bless \$self, $class;
}

sub equals {
	my ($self, $t) = @_;
	return $$self eq $$t;
}

sub promote {
	my ($self, $node, $type) = @_;

	return $node if $$self eq $$type;

	my ($self_dr, $type_dr);

	# Strip extra leading Ps (dereferences)
	($self_dr = $$self) =~ s/^(P*)//;
	if (($type_dr = $$type) !~ s/^$1//) {
		print "Cannot coerce array to scalar: $$self to $$type\n";
		return undef;
	}

	# We should strip prefix P*
	my $key = $self_dr . '<' . $type_dr;

	print "Key $key: ";

	my $name = ref($node);
	$name =~ s/^.*:://;

	if ($PROMOTIONS{$key}) {
		print "Promote $name from $$self to $$type\n";
	}
	else {
		print "Fail to promote $name from $$self to $$type\n";
		return undef;
	}

	return $node;
}

sub meet {
	my ($self, $other) = @_;
	return $self unless $other;

#	my $re = join("(?:", split(//, $$other));
#	$re = "((?:" . $re . (")?" x length $$other) . ")";
#
#	# print "$re\n";
#	$$self =~ /$re/;
#	my $match = $1;

	# This is cool.
	(my $t = $$self ^ $$other) =~ /^\0*/g;
	my $match = substr($$self, 0, pos($t));

	$match .= "_" if $match =~ /^P*$/;

	print "Match $$self =~ $$other: $match\n";

	return bless \$match, ref($self);

	return $self;
}

sub classp { return ${$_[0]} =~ /^S/; }
sub arrayp { return ${$_[0]} =~ /^P/; }

sub pointer {
	my ($self, $stars) = @_;	# $type is a Type
	$stars or $stars = 1;
	my $type = ("P" x $stars) . $$self;
	return bless \$type, ref($self);
}

sub deref {
	my ($self) = @_;
	my $class = __PACKAGE__;

	if ($$self eq '_') {
		my $out = "_";
		return bless \$out, $class;
	}
	elsif ($$self eq 's') {
		return T_INTEGER;
	}
	elsif ($$self =~ /^P/) {
		my $out = substr($$self, 1);
		# print "Array deref $$self to $out : $class\n";
		return bless \$out, $class;
	}
	confess "Can't deref type $$self";
}

sub dump { return ${$_[0]}; }

1;
